@@ -1,21 +1,29 @@
 """a collection of helper-functions to generate map-plots"""
 
-from functools import partial, lru_cache
+from functools import partial, lru_cache, wraps
 import warnings
+import copy
 
 import numpy as np
 import pandas as pd
-import geopandas as gpd
+
+try:
+    import geopandas as gpd
+except:
+    print("geopandas could not be imported... 'add_overlay' not working!")
+    pass
 from scipy.spatial import cKDTree
 from pyproj import CRS, Transformer
 
 import matplotlib as mpl
 import matplotlib.pyplot as plt
 from matplotlib import cm, collections
+from matplotlib.tri import Triangulation, TriMesh
 from matplotlib.colors import LinearSegmentedColormap
 from matplotlib.gridspec import GridSpec, GridSpecFromSubplotSpec, SubplotSpec
 from matplotlib.patches import Patch
 
+
 from cartopy import crs as ccrs
 from cartopy import feature as cfeature
 from cartopy.io import shapereader
@@ -90,17 +98,21 @@ class Maps(object):
 
         self._attached_cbs = dict()  # dict to memorize attached callbacks
 
+        self._cb = callbacks(self)
+
     def copy(
         self,
         data_specs=None,
         plot_specs=None,
         classify_specs=None,
+        copy_data=False,
     ):
         """
-        create a copy of the class that inherits all specifications
-        from the parent class (already loaded data is not copied!)
+        create a (deep)copy of the class that inherits all specifications
+        from the parent class.
+        Already loaded data is only copied if `copy_data=True`!
 
-        -> useful to quickly create plots with similar configuration but different data
+        -> useful to quickly create plots with similar configurations
 
         Parameters
         ----------
@@ -114,9 +126,15 @@ class Maps(object):
             a new Maps class.
         """
         initdict = dict()
-        initdict["data_specs"] = {**self.data_specs}
-        initdict["plot_specs"] = {**self.plot_specs}
-        initdict["classify_specs"] = {**self.classify_specs}
+        initdict["data_specs"] = {
+            key: copy.deepcopy(val) for key, val in self.data_specs.items()
+        }
+        initdict["plot_specs"] = {
+            key: copy.deepcopy(val) for key, val in self.plot_specs.items()
+        }
+        initdict["classify_specs"] = {
+            key: copy.deepcopy(val) for key, val in self.classify_specs.items()
+        }
 
         if data_specs:
             assert isinstance(data_specs, dict), "'data_specs' must be a dict"
@@ -137,15 +155,17 @@ class Maps(object):
         copy_cls.set_plot_specs(**initdict["plot_specs"])
         copy_cls.set_classify_specs(**initdict["classify_specs"])
 
+        if copy_data:
+            copy_cls.data = self.data.copy(deep=True)
+
         return copy_cls
 
     @property
-    @lru_cache()
     def cb(self):
         """
         accessor to pre-defined callback functions
         """
-        return callbacks(self)
+        return self._cb
 
     @property
     def data_specs(self):
@@ -259,8 +279,16 @@ class Maps(object):
             probability-density (True) or the number of counts per bin (False)
             The default is False.
         shape : str, optional
-            Indicator if "rectangles" or "ellipses" should be potted
-            The default is "ellipses".
+            Indicator how the data-points should be plotted
+                - "ellipses": plot projected ellipses
+                - "rectangles": plot projected rectangles
+                  Useful if pixel-boundaries should be visible.
+                  (Note: polygons will have a visible boundary even if the
+                   linewidth is set to 0!)
+                - "trimesh_rectangles": rectangles but drawn with a
+                  TriMesh collection so that there are no boundaries between
+                  the pixels. (e.g. useful for contourplots)
+            The default is "trimesh_rectangles".
         """
 
         for key, val in kwargs.items():
@@ -309,6 +337,7 @@ class Maps(object):
     def set_orientation(self, orientation="horizontal"):
         """
         set the orientation of the plot
+        (must be called before m.plot_map() !)
 
         Parameters
         ----------
@@ -331,6 +360,13 @@ class Maps(object):
             The instances must be provided as:  [matplotlib.figure, matplotlib.GridSpec]
             The default is None in which case a new figure is created.
         """
+
+        assert not hasattr(self, "figure"), (
+            "There is already an open figure! "
+            + "Either close it before creating a new one or call "
+            + "`m2 = m.copy()` to copy the Maps object and then use `m2.plot_map()"
+        )
+
         if not hasattr(self, "data"):
             print("you must set the data first!")
 
@@ -338,6 +374,18 @@ class Maps(object):
             data=self.data, **self.plot_specs, **self.data_specs, f_gridspec=f_gridspec
         )
 
+        # call the cleanup function if the figure is closed
+        # to ensure callbacks are removed
+        def on_close(event):
+            while len(self._attached_cbs) > 0:
+                self.remove_callback(list(self._attached_cbs)[-1])
+
+            # remove all figure properties
+            if hasattr(self, "figure"):
+                del self.figure
+
+        self.figure.f.canvas.mpl_connect("close_event", on_close)
+
     def _spatial_plot(
         self,
         data,
@@ -427,7 +475,10 @@ class Maps(object):
             or the probability density (True)
             The default is False.
         shape : str
-            the shapes to plot (either "ellipses" or "rectangles")
+            the shapes to plot
+                - ellipses
+                - rectangles
+                - trimesh_rectangles
         Returns
         -------
         dict :
@@ -448,12 +499,15 @@ class Maps(object):
             plot_epsg=plot_epsg,
             radius=radius,
             radius_crs=radius_crs,
-            cpos="c",
+            cpos=cpos,
             parameter=parameter,
             xcoord=xcoord,
             ycoord=ycoord,
         )
 
+        # remember props for later use
+        self._props = props
+
         if label is None:
             label = parameter
         if title is None:
@@ -554,7 +608,10 @@ class Maps(object):
 
         coll.set_picker(picker)
 
-        self.updatedict = dict(
+        # trigger drawing the figure
+        f.canvas.draw()
+
+        self.figure = _Maps_plot(
             f=f,
             gridspec=gs,
             colorbar_gs=cbgs,
@@ -572,8 +629,6 @@ class Maps(object):
             tick_precision=tick_precision,
         )
 
-        self.figure = _Maps_plot(**self.updatedict)
-
     def _prepare_data(
         self,
         data=None,
@@ -589,13 +644,14 @@ class Maps(object):
     ):
 
         # get specifications
-
         if data is None:
             data = self.data
         if xcoord is None:
             xcoord = self.data_specs["xcoord"]
         if ycoord is None:
             ycoord = self.data_specs["ycoord"]
+        if parameter is None:
+            parameter = self.data_specs["parameter"]
 
         if in_crs is None:
             in_crs = self.data_specs["in_crs"]
@@ -607,7 +663,7 @@ class Maps(object):
         if radius_crs is None:
             radius_crs = self.plot_specs["radius_crs"]
         if cpos is None:
-            cpos = "c"
+            cpos = self.plot_specs["cpos"]
         if shape is None:
             shape = self.plot_specs["shape"]
 
@@ -619,43 +675,65 @@ class Maps(object):
 
         # get coordinate transformation
         transformer = Transformer.from_crs(
-            CRS.from_user_input(in_crs), CRS.from_user_input(plot_epsg), always_xy=True
+            CRS.from_user_input(in_crs),
+            CRS.from_user_input(plot_epsg),
+            always_xy=True,
         )
 
+        # estimate radius if provided in "in_crs"
+        if (radius == "estimate") and (radius_crs == "in"):
+            radiusx = np.abs(np.diff(np.unique(xorig)).mean()) / 2.0
+            radiusy = np.abs(np.diff(np.unique(yorig)).mean()) / 2.0
+        # ... or get manually specified radius
+        elif isinstance(radius, (list, tuple)):
+            radiusx, radiusy = radius
+        else:
+            radiusx = radius
+            radiusy = radius
+
+        # fix position of pixel-center if radius is in "in_crs"
+        if radius_crs == "in":
+            if cpos == "c":
+                pass
+            elif cpos == "ll":
+                xorig += radiusx
+                yorig += radiusy
+            elif cpos == "ul":
+                xorig += radiusx
+                yorig -= radiusy
+            elif cpos == "lr":
+                xorig -= radiusx
+                yorig += radiusy
+            elif cpos == "ur":
+                xorig -= radiusx
+                yorig -= radiusx
+
+        # transform center-points
+        x0, y0 = transformer.transform(xorig, yorig)
+
+        # estimate radius if provided in "plot_epsg"
+        if (radius == "estimate") and (radius_crs == "out"):
+            radiusx = np.abs(np.diff(np.unique(x0)).mean()) / 2.0
+            radiusy = np.abs(np.diff(np.unique(y0)).mean()) / 2.0
+
+        # fix position of pixel-center if radius is in "plot_epsg"
+        if radius_crs == "out":
+            if cpos == "c":
+                pass
+            elif cpos == "ll":
+                x0 += radiusx
+                y0 += radiusy
+            elif cpos == "ul":
+                x0 += radiusx
+                y0 -= radiusy
+            elif cpos == "lr":
+                x0 -= radiusx
+                y0 += radiusy
+            elif cpos == "ur":
+                x0 -= radiusx
+                y0 -= radiusx
+
         if shape == "ellipses":
-            # transform center-points
-            x0, y0 = transformer.transform(xorig, yorig)
-
-            if radius == "estimate":
-                if radius_crs == "in":
-                    radiusx = np.abs(np.diff(np.unique(xorig)).mean()) / 2.0
-                    radiusy = np.abs(np.diff(np.unique(yorig)).mean()) / 2.0
-                elif radius_crs == "out":
-                    radiusx = np.abs(np.diff(np.unique(x0)).mean()) / 2.0
-                    radiusy = np.abs(np.diff(np.unique(y0)).mean()) / 2.0
-            elif isinstance(radius, (list, tuple)):
-                radiusx, radiusy = radius
-            else:
-                radiusx = radius
-                radiusy = radius
-
-            if radius_crs == "in":
-                # fix position of pixel-center
-                if cpos == "c":
-                    pass
-                elif cpos == "ll":
-                    xorig += radiusx
-                    yorig += radiusy
-                elif cpos == "ul":
-                    xorig += radiusx
-                    yorig -= radiusy
-                elif cpos == "lr":
-                    xorig += radiusx
-                    yorig -= radiusy
-                elif cpos == "ur":
-                    xorig -= radiusx
-                    yorig -= radiusx
-
             # transform corner-points
             if radius_crs == "in":
                 x3, y3 = transformer.transform(xorig + radiusx, yorig)
@@ -684,58 +762,8 @@ class Maps(object):
 
             theta = np.rad2deg(np.arcsin(np.abs(y3 - y0) / w))
 
-            if radius_crs == "out":
-                # fix position of pixel-center
-                if cpos == "c":
-                    pass
-                elif cpos == "ll":
-                    x0 += radiusx
-                    y0 += radiusy
-                elif cpos == "ul":
-                    x0 += radiusx
-                    y0 -= radiusy
-                elif cpos == "lr":
-                    x0 += radiusx
-                    y0 -= radiusy
-                elif cpos == "ur":
-                    x0 -= radiusx
-                    y0 -= radiusx
-
             props = dict(x0=x0, y0=y0, w=w, h=h, theta=theta, ids=ids, z_data=z_data)
         elif shape == "rectangles":
-            # transform center-points
-            x0, y0 = transformer.transform(xorig, yorig)
-
-            if radius == "estimate":
-                if radius_crs == "in":
-                    radiusx = np.abs(np.diff(np.unique(xorig)).mean()) / 2.0
-                    radiusy = np.abs(np.diff(np.unique(yorig)).mean()) / 2.0
-                elif radius_crs == "out":
-                    radiusx = np.abs(np.diff(np.unique(x0)).mean()) / 2.0
-                    radiusy = np.abs(np.diff(np.unique(y0)).mean()) / 2.0
-            elif isinstance(radius, (list, tuple)):
-                radiusx, radiusy = radius
-            else:
-                radiusx = radius
-                radiusy = radius
-
-            if radius_crs == "in":
-                # fix position of pixel-center
-                if cpos == "c":
-                    pass
-                elif cpos == "ll":
-                    xorig += radiusx
-                    yorig += radiusy
-                elif cpos == "ul":
-                    xorig += radiusx
-                    yorig -= radiusy
-                elif cpos == "lr":
-                    xorig += radiusx
-                    yorig -= radiusy
-                elif cpos == "ur":
-                    xorig -= radiusx
-                    yorig -= radiusx
-
             # transform corner-points
             if radius_crs == "in":
                 # top right
@@ -775,23 +803,6 @@ class Maps(object):
                 # bottom right
                 p3 = radius_t_p.transform(x0r + radiusx, y0r - radiusy)
 
-            if radius_crs == "out":
-                # fix position of pixel-center
-                if cpos == "c":
-                    pass
-                elif cpos == "ll":
-                    x0 += radiusx
-                    y0 += radiusy
-                elif cpos == "ul":
-                    x0 += radiusx
-                    y0 -= radiusy
-                elif cpos == "lr":
-                    x0 += radiusx
-                    y0 -= radiusy
-                elif cpos == "ur":
-                    x0 -= radiusx
-                    y0 -= radiusx
-
             # also attach max w & h (used for the kd-tree)
             props = dict(
                 verts=np.array(list(zip(*[np.array(i).T for i in (p0, p1, p2, p3)]))),
@@ -799,9 +810,89 @@ class Maps(object):
                 y0=y0,
                 ids=ids,
                 z_data=z_data,
-                w=(p0[0] - p1[0]).max(),
-                h=(p0[1] - p3[1]).max(),
+                w=(p0[0] - p1[0]),
+                h=(p0[1] - p3[1]),
             )
+
+        elif shape == "trimesh_rectangles":
+            # transform corner-points
+            if radius_crs == "in":
+                # top right
+                p0 = transformer.transform(xorig + radiusx, yorig + radiusy)
+                # top left
+                p1 = transformer.transform(xorig - radiusx, yorig + radiusy)
+                # bottom left
+                p2 = transformer.transform(xorig - radiusx, yorig - radiusy)
+                # bottom right
+                p3 = transformer.transform(xorig + radiusx, yorig - radiusy)
+
+            elif radius_crs == "out":
+                p0 = xorig + radiusx, yorig + radiusy
+                p1 = xorig - radiusx, yorig + radiusy
+                p2 = xorig - radiusx, yorig - radiusy
+                p3 = xorig + radiusx, yorig - radiusy
+            else:
+                radius_t = Transformer.from_crs(
+                    CRS.from_user_input(in_crs),
+                    CRS.from_user_input(radius_crs),
+                    always_xy=True,
+                )
+                radius_t_p = Transformer.from_crs(
+                    CRS.from_user_input(radius_crs),
+                    CRS.from_user_input(plot_epsg),
+                    always_xy=True,
+                )
+
+                x0r, y0r = radius_t.transform(xorig, yorig)
+
+                # top right
+                p0 = radius_t_p.transform(x0r + radiusx, y0r + radiusy)
+                # top left
+                p1 = radius_t_p.transform(x0r - radiusx, y0r + radiusy)
+                # bottom left
+                p2 = radius_t_p.transform(x0r - radiusx, y0r - radiusy)
+                # bottom right
+                p3 = radius_t_p.transform(x0r + radiusx, y0r - radiusy)
+
+            verts = np.array(list(zip(*[np.array(i).T for i in (p0, p1, p2, p3)])))
+            x = np.vstack(
+                [verts[:, 2][:, 0], verts[:, 3][:, 0], verts[:, 1][:, 0]]
+            ).T.ravel()
+            y = np.vstack(
+                [verts[:, 2][:, 1], verts[:, 3][:, 1], verts[:, 1][:, 1]]
+            ).T.ravel()
+
+            x2 = np.vstack(
+                [verts[:, 3][:, 0], verts[:, 0][:, 0], verts[:, 1][:, 0]]
+            ).T.ravel()
+            y2 = np.vstack(
+                [verts[:, 3][:, 1], verts[:, 0][:, 1], verts[:, 1][:, 1]]
+            ).T.ravel()
+
+            x = np.append(x, x2)
+            y = np.append(y, y2)
+
+            tri = Triangulation(
+                x, y, triangles=np.array(range(len(x))).reshape((len(x) // 3, 3))
+            )
+
+            # also attach max w & h (used for the kd-tree)
+            props = dict(
+                tri=tri,
+                x0=x0,
+                y0=y0,
+                ids=ids,
+                z_data=z_data,
+                w=(p0[0] - p1[0]),
+                h=(p0[1] - p3[1]),
+            )
+
+        else:
+            raise TypeError(
+                f"'{shape}' is not a valid shape, use one of:\n"
+                + "    - 'ellipses'\n    - 'rectangles'\n    - 'trimesh_rectangles'"
+            )
+
         return props
 
     def _classify_data(self, z_data, cmap, histbins, vmin, vmax, classify_specs=None):
@@ -958,7 +1049,17 @@ class Maps(object):
                 transOffset=ax.transData,
             )
 
-        if shape == "rectangles":
+            if color is not None:
+                coll.set_color(color)
+            else:
+                coll.set_array(np.ma.masked_invalid(z_data))
+                coll.set_cmap(cmap)
+                coll.set_clim(vmin, vmax)
+                coll.set_norm(norm)
+            # coll.set_urls(ids)
+            ax.add_collection(coll)
+
+        elif shape == "rectangles":
             coll = collections.PolyCollection(
                 verts=props["verts"],
                 transOffset=ax.transData,
@@ -966,18 +1067,36 @@ class Maps(object):
             # add centroid positions (used by the picker in self._spatial_plot)
             coll._Maps_positions = list(zip(props["x0"], props["y0"]))
 
-        if color is not None:
-            coll.set_color(color)
-        else:
-            coll.set_array(np.ma.masked_invalid(z_data))
-            coll.set_cmap(cmap)
-            coll.set_clim(vmin, vmax)
-            coll.set_norm(norm)
-        coll.set_urls(ids)
+            if color is not None:
+                coll.set_color(color)
+            else:
+                coll.set_array(np.ma.masked_invalid(z_data))
+                coll.set_cmap(cmap)
+                coll.set_clim(vmin, vmax)
+                coll.set_norm(norm)
+            # coll.set_urls(ids)
+            ax.add_collection(coll)
 
-        # coll.set_facecolor(cbcmap(z_data))    # do this to properly treat nan-values
-        # coll.set_edgecolor('none')
-        ax.add_collection(coll)
+        elif shape == "trimesh_rectangles":
+            coll = TriMesh(props["tri"])
+            # add centroid positions (used by the picker in self._spatial_plot)
+            coll._Maps_positions = list(zip(props["x0"], props["y0"]))
+
+            if color is not None:
+                coll.set_facecolors([color] * (len(props["x0"])) * 6)
+            else:
+                z = np.ma.masked_invalid(z_data)
+                # tri-contour meshes need 3 values for each triangle
+                z = np.broadcast_to(z, (3, len(z))).T
+                # we plot 2 triangles per rectangle
+                z = np.broadcast_to(z, (2, *z.shape))
+
+                coll.set_array(z.ravel())
+                coll.set_cmap(cmap)
+                coll.set_clim(vmin, vmax)
+                coll.set_norm(norm)
+            # coll.set_urls(np.repeat(ids, 3, axis=0))
+            ax.add_collection(coll)
 
         return coll
 
@@ -1169,12 +1288,16 @@ class Maps(object):
         **kwargs :
             all kwargs are passed to `gdf.plot(**kwargs)`
         """
-        ax = self.updatedict["ax"]
+        assert hasattr(self, "figure"), "you must call .plot_map() first!"
+
+        ax = self.figure.ax
 
         defaultargs = dict(facecolor="none", edgecolor="k", lw=1.5)
         defaultargs.update(kwargs)
 
-        gdf.plot(ax=ax, aspect=ax.get_aspect(), **defaultargs)
+        gdf.to_crs(epsg=self.plot_specs["plot_epsg"]).plot(
+            ax=ax, aspect=ax.get_aspect(), **defaultargs
+        )
 
     def add_overlay(
         self,
@@ -1221,9 +1344,9 @@ class Maps(object):
         if legendlabel is None:
             legendlabel = dataspec.get("name", "overlay")
 
-        assert hasattr(self, "updatedict"), "you must call .plot_map() first!"
+        assert hasattr(self, "figure"), "you must call .plot_map() first!"
 
-        ax = self.updatedict["ax"]
+        ax = self.figure.ax
 
         if not all([i in dataspec for i in ["resolution", "category", "name"]]):
             assert False, (
@@ -1271,13 +1394,15 @@ class Maps(object):
         vmax=None,
         color=None,
         radius="estimate",
-        radius_crs="in",
-        in_crs=4326,
-        cpos="c",
+        radius_crs=None,
+        in_crs=None,
+        cpos=None,
         legend_kwargs=True,
         shape="ellipses",
     ):
         """
+        add another layer of pixels
+
         Parameters
         ----------
         data : pandas.DataFrame
@@ -1323,6 +1448,7 @@ class Maps(object):
         shape : str
             the shapes to plot (either "ellipses" or "rectangles")
         """
+        assert hasattr(self, "figure"), "you must call .plot_map() first!"
 
         if parameter is None:
             parameter = next(i for i in data.keys() if i not in [xcoord, ycoord])
@@ -1437,56 +1563,219 @@ class Maps(object):
             assert hasattr(self.cb, callback), (
                 f"The function '{callback}' does not exist as a pre-defined callback."
                 + " Use one of:\n    - "
-                + "\n    - ".join(
-                    [i for i in self.cb.__dir__() if not i.startswith("_")]
-                )
+                + "\n    - ".join(self.cb.cb_list)
             )
             callback = getattr(self.cb, callback)
+        elif callable(callback):
+            # re-bind the callback methods to the eomaps.Maps.cb object
+            # in case custom functions are used
+            if hasattr(callback, "__func__"):
+                callback = callback.__func__.__get__(self.cb)
+            else:
+                callback = callback.__get__(self.cb)
+
+        # add mouse-button assignment as suffix to the name (with __ separator)
+        # TODO document this!
+        cbname = callback.__name__ + f"__{double_click}_{mouse_button}"
 
         assert (
-            callback.__name__ not in self._attached_cbs
-        ), f"the callback '{callback.__name__}' is already attached to the plot!"
+            cbname not in self._attached_cbs
+        ), f"the callback '{cbname}' is already attached to the plot!"
 
-        # re-bind the callback methods to the eomaps.Maps object
-        # in case custom functions are used
-        if hasattr(callback, "__func__"):
-            callback = callback.__func__.__get__(self)
-        else:
-            newcb = partial(callback, self=self)
-            newcb.__name__ = callback.__name__
-            callback = newcb
+        # TODO support multiple assignments for callbacks
+        # make sure multiple callbacks of the same funciton are only assigned
+        # if multiple assignments are properly handled
+        multi_cb_functions = ["mark", "annotate"]
+
+        no_multi_cb = [*self.cb.cb_list]
+        for i in multi_cb_functions:
+            no_multi_cb.pop(no_multi_cb.index(i))
+
+        if callback.__name__ in no_multi_cb:
+            assert callback.__name__ not in [
+                i.split("__")[0] for i in self._attached_cbs
+            ], (
+                "Multiple assignments of the callback"
+                + f" '{callback.__name__}' are not (yet) supported..."
+            )
 
         # ------------- add a callback
         def onpick(event):
-            print(event.double_click, event.mouse_button)
-
             if (event.double_click == double_click) and (
                 event.mouse_button == mouse_button
             ):
                 ind = event.ind
                 if ind is not None:
-                    if isinstance(event.artist, collections.EllipseCollection):
+                    if isinstance(
+                        event.artist,
+                        (
+                            collections.EllipseCollection,
+                            collections.PolyCollection,
+                            collections.TriMesh,
+                        ),
+                    ):
                         clickdict = dict(
-                            pos=self.figure.coll.get_offsets()[ind],
-                            ID=self.figure.coll.get_urls()[ind],
-                            val=self.figure.coll.get_array()[ind],
+                            pos=(self._props["x0"][ind], self._props["y0"][ind]),
+                            ID=self._props["ids"][ind],
+                            val=self._props["z_data"][ind],
+                            ind=ind,
                         )
 
-                        callback(**clickdict, **kwargs)
-                    elif isinstance(event.artist, collections.PolyCollection):
-                        clickdict = dict(
-                            pos=self.figure.coll._Maps_positions[ind],
-                            ID=self.figure.coll.get_urls()[ind],
-                            val=self.figure.coll.get_array()[ind],
-                        )
+                    callback(**clickdict, **kwargs)
 
-                        callback(**clickdict, **kwargs)
                 else:
-                    if "annotate" in self._attached_cbs:
+                    if "annotate" in [i.split("__")[0] for i in self._attached_cbs]:
                         self._cb_hide_annotate()
 
-        self._attached_cbs[callback.__name__] = self.figure.f.canvas.mpl_connect(
-            "pick_event", onpick
+        cid = self.figure.f.canvas.mpl_connect("pick_event", onpick)
+        self._attached_cbs[cbname] = cid
+
+        return cid
+
+    def add_marker(
+        self,
+        ID=None,
+        xy=None,
+        xy_crs=None,
+        radius=None,
+        shape="circle",
+        buffer=1,
+        **kwargs,
+    ):
+        """
+        add a marker to the plot
+
+        Parameters
+        ----------
+        ID : any
+            The index-value of the pixel in m.data.
+        xy : tuple
+            A tuple of the position of the pixel provided in "xy_crs".
+            If None, xy must be provided in the coordinate-system of the plot!
+            The default is None
+        xy_crs : any
+            the identifier of the coordinate-system for the xy-coordinates
+        radius : float or None, optional
+            The radius of the marker. If None, it will be evaluated based
+            on the pixel-spacing of the provided dataset
+            The default is None.
+        shape : str, optional
+            Indicator which shape to draw. Currently supported shapes are:
+                - circle
+                - ellipse
+                - rectangle
+
+            The default is "circle".
+        buffer : float, optional
+            A factor to scale the size of the shape. The default is 1.
+        **kwargs :
+            kwargs passed to the matplotlib patch.
+            (e.g. `facecolor`, `edgecolor`, `linewidth`, `alpha` etc.)
+
+        Returns
+        -------
+        None.
+
+        """
+        if ID is not None:
+            assert xy is None, "You can only provide 'ID' or 'pos' not both!"
+
+            xy = self.data.loc[ID][
+                [self.data_specs["xcoord"], self.data_specs["ycoord"]]
+            ].values
+            xy_crs = self.data_specs["in_crs"]
+
+            ind = self.data.index.get_loc(ID)
+
+        if xy is not None:
+            ind = None
+            if xy_crs is not None:
+                # get coordinate transformation
+                transformer = Transformer.from_crs(
+                    CRS.from_user_input(xy_crs),
+                    CRS.from_user_input(self.plot_specs["plot_epsg"]),
+                    always_xy=True,
+                )
+                # transform coordinates
+                xy = transformer.transform(*xy)
+
+        # add marker
+        self.cb.mark(
+            ID=ID, pos=xy, radius=radius, ind=ind, shape=shape, buffer=buffer, **kwargs
+        )
+
+    def add_annotation(
+        self,
+        ID=None,
+        xy=None,
+        xy_crs=None,
+        text=None,
+        **kwargs,
+    ):
+        """
+        add an annotation to the plot
+
+        Parameters
+        ----------
+        ID : any
+            The index-value of the pixel in m.data.
+        xy : tuple
+            A tuple of the position of the pixel provided in "xy_crs".
+            If None, xy must be provided in the coordinate-system of the plot!
+            The default is None
+        xy_crs : any
+            the identifier of the coordinate-system for the xy-coordinates
+        text : callable or str, optional
+            if str: the string to print
+            if callable: A function that returns the string that should be
+            printed in the annotation with the following call-signature:
+
+                >>> def text(m, ID, val, pos):
+                >>>     # m   ... the Maps object
+                >>>     # ID  ... the ID
+                >>>     # pos ... the position
+                >>>     # val ... the value
+                >>>
+                >>>     return "the string to print"
+
+            The default is None.
+
+        **kwargs
+            kwargs passed to m.cb.annotate
+        Returns
+        -------
+        None.
+
+        """
+        if ID is not None:
+            assert xy is None, "You can only provide 'ID' or 'pos' not both!"
+
+            xy = self.data.loc[ID][
+                [self.data_specs["xcoord"], self.data_specs["ycoord"]]
+            ].values
+            xy_crs = self.data_specs["in_crs"]
+
+        if xy is not None:
+
+            if xy_crs is not None:
+                # get coordinate transformation
+                transformer = Transformer.from_crs(
+                    CRS.from_user_input(xy_crs),
+                    CRS.from_user_input(self.plot_specs["plot_epsg"]),
+                    always_xy=True,
+                )
+                # transform coordinates
+                xy = transformer.transform(*xy)
+
+        # add marker
+        self.cb.annotate(
+            ID=ID,
+            pos=xy,
+            val=None if ID is None else self.data.loc[ID][self.data_specs["parameter"]],
+            ind=None if ID is None else self.data.index.get_loc(ID),
+            permanent=True,
+            text=text,
+            **kwargs,
         )
 
     def remove_callback(self, callback):
@@ -1495,28 +1784,69 @@ class Maps(object):
 
         Parameters
         ----------
-        callback : callable or string
+        callback : int, str or callable
+            if int: the identification-number of the callback to remove
+                    (the number is returned by `cid = m.add_callback()`)
+            if str: the name of the callback to remove
+                    (format: `<function_name>__<double_click>_<mouse_button>`)
+            if callable: the `__name__` property of the callback is used to
+                         remove ANY callback that references the corresponding
+                         function
+
             either the callback-function that should be removed from the figure
             (or the name of the function)
         """
-        if isinstance(callback, str):
-            name = callback
+
+        if isinstance(callback, int):
+            self.figure.f.canvas.mpl_disconnect(callback)
+            name = dict(zip(self._attached_cbs.values(), self._attached_cbs.keys()))[
+                callback
+            ]
+            del self._attached_cbs[name]
+
+            # call cleanup methods on removal
+            fname = name.split("__")[0]
+            if hasattr(self.cb, f"_{fname}_cleanup"):
+                getattr(self.cb, f"_{fname}_cleanup")(self)
+
+            print(f"Removed the callback: '{name}'.")
+
         else:
-            name = callback.__name__
+            if isinstance(callback, str):
+                names = [callback]
+                if names[0] not in self._attached_cbs:
+                    warnings.warn(
+                        f"The callback '{names[0]}' is not attached and can not"
+                        + " be removed. Attached callbacks are:\n    - "
+                        + "    - \n".join(list(self._attached_cbs))
+                    )
+                    return
 
-        if name not in self._attached_cbs:
-            warnings.warn(
-                f"The callback '{name}' is not attached and can not"
-                + " be removed. Attached callbacks are:\n    - "
-                + "    - \n".join(list(self._attached_cbs))
-            )
+            elif callable(callback):
+                # identify all callbacks that relate to the provided function
+                names = [
+                    key
+                    for key in self._attached_cbs
+                    if key.split("__")[0] == callback.__name__
+                ]
+                if len(names) == 0:
+                    warnings.warn(
+                        f"The callback '{callback.__name__}' is not attached"
+                        + "and can not be removed. Attached callbacks are:"
+                        + "\n    - "
+                        + "    - \n".join(list(self._attached_cbs))
+                    )
 
-        self.figure.f.canvas.mpl_disconnect(self._attached_cbs[name])
-        del self._attached_cbs[name]
+            for name in names:
+                self.figure.f.canvas.mpl_disconnect(self._attached_cbs[name])
+                del self._attached_cbs[name]
 
-        # call cleanup methods on removal
-        if hasattr(self.cb, f"_{name}_cleanup"):
-            getattr(self.cb, f"_{name}_cleanup")(self)
+                # call cleanup methods on removal
+                fname = name.split("__")[0]
+                if hasattr(self.cb, f"_{fname}_cleanup"):
+                    getattr(self.cb, f"_{fname}_cleanup")()
+
+                print(f"Removed the callback: '{name}'.")
 
     @lru_cache()
     def _get_crs(self, crs):
@@ -1524,38 +1854,45 @@ class Maps(object):
 
     def _cb_hide_annotate(self):
         # a function to hide the annotation of an empty area is clicked
-        if hasattr(self, "annotation"):
-            self.annotation.set_visible(False)
-            self._blit()
+        if hasattr(self.cb, "annotation"):
+            self.cb.annotation.set_visible(False)
+            self._blit(self.cb.annotation)
 
-    # implement blitting
-    # https://stackoverflow.com/a/29284318/9703451
+    # implement blitting (see https://stackoverflow.com/a/29284318/9703451)
     def _safe_draw(self):
         """Temporarily disconnect the draw_event callback to avoid recursion"""
         canvas = self.figure.f.canvas
-        canvas.mpl_disconnect(self.draw_cid)
+        if hasattr(self, "draw_cid"):
+            canvas.mpl_disconnect(self.draw_cid)
 
-        canvas.draw()
-        self.draw_cid = canvas.mpl_connect("draw_event", self._grab_background)
+            canvas.draw()
+            self.draw_cid = canvas.mpl_connect("draw_event", self._grab_background)
+        else:
+            canvas.draw()
 
-    def _grab_background(self, event=None):
+    def _grab_background(self, event=None, redraw=True):
         """
         When the figure is resized, draw everything, and update the background.
         """
+        # hide annotations from the background
         annotation_visible = False
-        if hasattr(self, "annotation"):
-            if self.annotation.get_visible():
+        if hasattr(self.cb, "annotation"):
+            if self.cb.annotation.get_visible():
                 annotation_visible = True
-                self.annotation.set_visible(False)
-        self._safe_draw()
+                self.cb.annotation.set_visible(False)
+
+        if redraw:
+            self._safe_draw()
 
         # With most backends (e.g. TkAgg), we could grab (and refresh, in
         # self.blit) self.ax.bbox instead of self.fig.bbox, but Qt4Agg, and
         # some others, requires us to update the _full_ canvas, instead.
         self.background = self.figure.f.canvas.copy_from_bbox(self.figure.f.bbox)
+
+        # re-draw annotations
         if annotation_visible:
-            self.annotation.set_visible(True)
-            self._blit(self.annotation)
+            self.cb.annotation.set_visible(True)
+            self._blit(self.cb.annotation)
         else:
             self._blit()
 
@@ -1572,3 +1909,7 @@ class Maps(object):
         if artist is not None:
             self.figure.ax.draw_artist(artist)
         self.figure.f.canvas.blit(self.figure.f.bbox)
+
+    @wraps(plt.savefig)
+    def savefig(self, *args, **kwargs):
+        self.figure.f.savefig(*args, **kwargs)
